/*
 * Copyright (C) 2011 Kenichi Takahashi
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

/* EGL を使用する際に必要なヘッダ */
#include <EGL/egl.h>
/* OpenGL ES を使用する際に必要なヘッダ */
#include <GLES/gl.h>

/* OpenSL ES を使用する際に必要なヘッダ */
#include <SLES/OpenSLES.h>
/* OpenSL ES の Android 拡張機能を使用する際に必要なヘッダ */
#include <SLES/OpenSLES_Android.h>

/* sensor 使用に必要 */
#include <android/sensor.h>

/*
 * native_activity.h をダイレクトに使っても良いのですが、android_native_app_glue.h
 * という helper ライブラリが用意されていますので、これを使うことにします
 * それにより Activity のライフサイクルの各イベントの処理の実装が簡単になります
 */
#include <android_native_app_glue.h>
#include <android/asset_manager.h>

#include <assert.h>
#include <errno.h>
#include <sys/time.h>
#include <time.h>
#include <android/log.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

/* ティーポットのモデルデータ */
#include "teapot.h"

#define  LOG_TAG    "SD8sample"

/* 3D ベクトル */
typedef struct {
    float x;  /* X 成分*/
    float y;  /* Y 成分*/
    float z;  /* Z 成分*/
} vec3f;

struct engine {
    struct android_app* app;

    /* センサ関連 */
    ASensorManager* sensorManager;
    const ASensor* accelerometerSensor;
    ASensorEventQueue* sensorEventQueue;

    int needToDraw; /* 描画の必要を示すフラグ */
    int playing; /* 音声再生状態 */
    /* タップ座標の保持 */
    float xpos; /* X座標 */
    float ypos; /* Y座標 */

    vec3f acc;  /* 加速度センサからの値 */

    /* EGL 関連 */
    EGLDisplay display;
    EGLSurface surface;
    EGLContext context;
    int32_t width;  /* 画面の幅 */
    int32_t height; /* 画面の高さ */
    float aspect;   /* アスペクト比 */

    float yrot;     /* Y軸の回転量（°） */
    float matrix[16]; /* 変換行列 */
};

/* engine interfaces */
static SLObjectItf engineObject = NULL;
static SLEngineItf engineEngine;

/* output mix interfaces */
static SLObjectItf outputMixObject = NULL;

/* file descriptor player interfaces */
static SLObjectItf fdPlayerObject = NULL;
static SLPlayItf fdPlayerPlay;
static SLSeekItf fdPlayerSeek;

static void setPlayingAssetAudioPlayer(struct engine *engine, int isPlaying);
static void playerCallback(SLPlayItf caller, void *context, SLuint32 event);

/*
 * 画面周り（EGL）の初期化
 */
static int engine_init_display(struct engine* engine) {
    /*
     * EGL のアトリビュートの定義
     * 各色（RGB）に 8 ビットの割り当て
     * デプス（奥行き）に 16 ビットの割り当て
     */
    const EGLint attribs[] = {
            EGL_SURFACE_TYPE, EGL_WINDOW_BIT,
            EGL_BLUE_SIZE, 8,
            EGL_GREEN_SIZE, 8,
            EGL_RED_SIZE, 8,
	    EGL_DEPTH_SIZE, 16,
            EGL_NONE
    };
    EGLint w, h, dummy, format;
    EGLint numConfigs;
    EGLConfig config;
    EGLSurface surface;
    EGLContext context;

    /* ディスプレイを取得 */
    EGLDisplay display = eglGetDisplay(EGL_DEFAULT_DISPLAY);

    /* 取得したディスプレイで EGL を初期化 */
    eglInitialize(display, 0, 0);

    /* 冒頭で設定した定義でバッファ設定を選択します */
    eglChooseConfig(display, attribs, &config, 1, &numConfigs);

    /*
     * Android のネイティブウィンドウと関連づけるためのアトリビュートの ID を
     * format に取得します
     */
    eglGetConfigAttrib(display, config, EGL_NATIVE_VISUAL_ID, &format);

    /* Android のネイティブウィンドウと関連づけます */
    ANativeWindow_setBuffersGeometry(engine->app->window, 0, 0, format);

    /* サーフェスの生成 */
    surface = eglCreateWindowSurface(display, config, engine->app->window, NULL);
    /* コンテキストの生成 */
    context = eglCreateContext(display, config, NULL, NULL);

    /* 生成したコンテキストをカレントにします */
    if (eglMakeCurrent(display, surface, surface, context) == EGL_FALSE) {
        __android_log_print (ANDROID_LOG_INFO, LOG_TAG, "Unable to eglMakeCurrent");
        return -1;
    }

    /* サーフェスサイズの取得*/
    eglQuerySurface(display, surface, EGL_WIDTH, &w);  /* 画面の幅 */
    eglQuerySurface(display, surface, EGL_HEIGHT, &h); /* 画面の高さ */
    __android_log_print (ANDROID_LOG_INFO, LOG_TAG, "display surface w=%d h=%d", w, h);

    engine->display = display;
    engine->context = context;
    engine->surface = surface;
    engine->width = w;
    engine->height = h;

    /* 画面のアスペクト比（縦横比）を求める */
    engine->aspect = (float)h / (float)w;

    /* GL の初期状態を設定 */
    /* パースペクティブコレクションを速度優先にします */
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
    /* 裏面（見えていない側）は描画しない */
    glEnable(GL_CULL_FACE);
    /* 色の変化をスムースに */
    glShadeModel(GL_SMOOTH);
    /* デプステスト（奥行き方向の描画可否）を有効にします */
    glEnable(GL_DEPTH_TEST);

    /* GL の描画領域を設定（画面一杯に） */
    glViewport(0, 0, w, h);
    return 0;
}

/*
 * 4x4 の単位行列を設定
 */
static void setIdentityMatrix(float *m)
{
	m[0] = 1.0f;
	m[1] = 0.0f;
	m[2] = 0.0f;
	m[3] = 0.0f;
	
	m[4] = 0.0f;
	m[5] = 1.0f;
	m[6] = 0.0f;
	m[7] = 0.0f;
	
	m[8] = 0.0f;
	m[9] = 0.0f;
	m[10] = 1.0f;
	m[11] = 0.0f;	
	
	m[12] = 0.0f;
	m[13] = 0.0f;
	m[14] = 0.0f;
	m[15] = 1.0f;
}

/*
 * 3D ベクトルを正規化
 */
static void normalizeVector(vec3f *v)
{
  /* 大きさを求める */
  float length = sqrtf(v->x * v->x + v->y * v->y + v->z * v->z);
  /* 各成分を大きさで割る（大きさが 1 になるようにします） */
  v->x /= length;
  v->y /= length;
  v->z /= length;
}

/*
 * 加速度センサの値から回転行列の値を計算
 */
static void getRotationMatrix(struct engine* engine)
{
    vec3f xaxis; /* X軸を示すベクトル */
    vec3f yaxis; /* Y軸を示すベクトル */
    vec3f zaxis; /* Z軸を示すベクトル */
    /*
     * 加速度センサからの値を重力の方向としてZ軸のベクトルとします
     * ベクトルの大きさを重力加速度とみなして正規化
     */
    zaxis.x = engine->acc.x;
    zaxis.y = engine->acc.y;
    zaxis.z = engine->acc.z;
    normalizeVector(&zaxis);

    /* Y 軸のベクトルは加速度センサの Y 方向と Z 方向から合成 */
    yaxis.x = 0;
    yaxis.y = zaxis.z;
    yaxis.z = -zaxis.y;
    /* 大きさが変わるので正規化が必要 */
    normalizeVector(&yaxis);

    /* X軸のベクトル = Z軸とY軸の外積 */
    xaxis.x = yaxis.y * zaxis.z - yaxis.z * zaxis.y;
    xaxis.y = yaxis.z * zaxis.x - yaxis.x * zaxis.z;
    xaxis.z = yaxis.x * zaxis.y - yaxis.y * zaxis.x;

    /* 4x4 の行列に各軸の成分を代入します */
    float *m = engine->matrix;
    /* X軸 */
    m[0] = xaxis.x;
    m[1] = xaxis.y;
    m[2] = xaxis.z;
    /* Y軸 */
    m[4] = yaxis.x;
    m[5] = yaxis.y;
    m[6] = yaxis.z;
    /* Z軸 */
    m[8] = zaxis.x;
    m[9] = zaxis.y;
    m[10] = zaxis.z;
}

/*
 * ティーポットの描画
 */
static void drawTeapot(struct engine* engine)
{
    static const float projectionWidth = 0.25f;
    /* プロジェクションモードの行列を設定します */
    glMatrixMode(GL_PROJECTION);
    /* 単位行列で初期化 */
    glLoadIdentity();
    /* 正射影による描画の設定 */
    glOrthof(-projectionWidth, projectionWidth,
	     -projectionWidth * engine->aspect, projectionWidth * engine->aspect,
	     -1.0f, 1.0f);
    /* モデルビューモードの行列を設定します */
    glMatrixMode(GL_MODELVIEW);
    /* 単位行列で初期化 */
    glLoadIdentity();
    /* 加速度センサの値から回転行列を得る */
    getRotationMatrix(engine);
    /* 取得した回転行列をロードします */
    glLoadMatrixf(engine->matrix);
    glRotatef(90, 1, 0, 0);
    /* 画面タップ操作による Y 軸の回転を設定します */
    glRotatef(engine->yrot, 0, 1, 0);

    /* 頂点データの設定 */
    /*
     * 頂点座標のポインタの設定
     * teapot.h の float teapot_vertices[] には X, Y, Z の頂点座標列が設定されています
     */
    glVertexPointer(3, GL_FLOAT, 0, teapot_vertices);
    glEnableClientState(GL_VERTEX_ARRAY);
    /*
     * 法線ベクトルのポインタの設定
     * teapot.h の float teapot_normals[] には各頂点毎の法線ベクトル(X, Y, Z の各成分)
     * の列が設定されています
     */
    glNormalPointer(GL_FLOAT, 0, teapot_normals);
    glEnableClientState(GL_NORMAL_ARRAY);

    int i;
    /*
     * 設定した頂点データで描画します
     * teapot.h の
     * num_teapot_indices ... トライアングルストリップで描画するインデックスの固まりの数
     * new_teapot_indicies[i] ... 固まりのインデックスの要素数
     * new_teapot_indicies[i+1] ... インデックスが要素数だけ連続して設定されています
     */
    for(i = 0; i < num_teapot_indices; i += new_teapot_indicies[i] + 1) {
        glDrawElements(GL_TRIANGLE_STRIP, new_teapot_indicies[i],
		       GL_UNSIGNED_SHORT, &new_teapot_indicies[i+1]);
    }
}

/*
 * 画面の描画を行う必要がある時に呼び出される関数
 */
static void engine_draw_frame(struct engine* engine)
{
    /* 環境光の色の定義 R, G, B, A */
    float global_ambient[] = { 1.0f, 1.0f, 1.0f, 1.0f };
    if (engine->app->window == NULL) {
        // No window.
        return;
    }

    /* 音声の再生状態で処理を切り替える */
    if (engine->playing) {
      /* 再生なら R 成分の値を 0 にして描画色を変える */
      global_ambient[0] = 0.0f;
    }

    /* バッファのクリア色を設定 R, G, B, A */
    glClearColor(0, 0, 0, 1); /* 黒、不透明 */
    /* カラーバッファとデプスバッファをクリアします */
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    /* ライティングを有効にします */
    glEnable(GL_LIGHTING);
    /* １つ目のライトを有効にします */
    glEnable(GL_LIGHT0);
    /* 環境光を設定します */
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, global_ambient);

    /* ティーポットの描画 */
    drawTeapot(engine);

    /* １つ目のライトを無効にします */
    glDisable(GL_LIGHT0);
    /* ライティングを無効にします */
    glDisable(GL_LIGHTING);

    /* スワップバッファ（見えているバッファの切替） */
    eglSwapBuffers(engine->display, engine->surface);
    /* 描画が終了したので描画フラグを OFF */
    engine->needToDraw = 0;
}

/*
 * アプリ終了時に呼び出される関数の定義です
 */
static int engine_term_display(struct engine* engine) {
    engine->needToDraw = 0;
    engine->playing = 0;

    /* OpenSL ES 関連のオブジェクトを解放します */
    /* プレイヤーオブジェクト */
    if (fdPlayerObject != NULL) {
        (*fdPlayerObject)->Destroy(fdPlayerObject);
        fdPlayerObject = NULL;
        fdPlayerPlay = NULL;
        fdPlayerSeek = NULL;
    }
    /* 出力先オブジェクト */
    if (outputMixObject != NULL) {
        (*outputMixObject)->Destroy(outputMixObject);
        outputMixObject = NULL;
    }
    /* エンジンオブジェクト */
    if (engineObject != NULL) {
        (*engineObject)->Destroy(engineObject);
        engineObject = NULL;
        engineEngine = NULL;
    }

    /* EGL 関連の終了処理 */
    if (engine->display != EGL_NO_DISPLAY) {
        /* コンテキストの解放 */
        eglMakeCurrent(engine->display, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT);
        if (engine->context != EGL_NO_CONTEXT) {
            eglDestroyContext(engine->display, engine->context);
        }
        /* サーフェスの解放 */
        if (engine->surface != EGL_NO_SURFACE) {
            eglDestroySurface(engine->display, engine->surface);
        }
        /* EGL の終了 */
        eglTerminate(engine->display);
    }
    engine->display = EGL_NO_DISPLAY;
    engine->context = EGL_NO_CONTEXT;
    engine->surface = EGL_NO_SURFACE;
}

/*
 * 入力イベントがあった時に呼び出される関数
 */
static int32_t engine_handle_input(struct android_app* app, AInputEvent* event)
{
    struct engine* engine = (struct engine*)app->userData;
    if (AInputEvent_getType(event) == AINPUT_EVENT_TYPE_MOTION) {
        /* 画面のタップ操作の場合 */
        __android_log_print (ANDROID_LOG_INFO, LOG_TAG,
		"Motion event: action=%d flag=%d metaState=0x%x",
                AMotionEvent_getAction(event),
                AMotionEvent_getFlags(event),
	        AMotionEvent_getMetaState(event));
	int32_t eventAction = AMotionEvent_getAction(event);
	if (eventAction == AMOTION_EVENT_ACTION_DOWN) {
          /* タップ操作が画面に触れた状態の場合 */
	  /* タップ位置の取得 */
	  engine->xpos = AMotionEvent_getX(event, 0); /* X座標 */
	  engine->ypos = AMotionEvent_getY(event, 0); /* Y座標 */
	  __android_log_print (ANDROID_LOG_INFO, LOG_TAG,
			       "pos = (%f, %f)", 
			       engine->xpos, engine->ypos);
	} else if (eventAction == AMOTION_EVENT_ACTION_UP) {
          /* タップ操作の指が画面から離れた時 */
	  /* タップ位置の取得 */
	  float x = AMotionEvent_getX(event, 0); /* X座標 */
	  float y = AMotionEvent_getY(event, 0); /* Y座標 */
	  __android_log_print (ANDROID_LOG_INFO, LOG_TAG,
			       "up = (%f, %f)", 
			       x, y);
	  if (x == engine->xpos && y == engine->ypos) {
              /*
               * 指が触れた時と離れた時の座標に差がなければタップ操作を有効
	       * としてプレイヤーを再生状態にします
               */
	      setPlayingAssetAudioPlayer(engine, 1);
	      /* 描画必要フラグをON */
	      engine->needToDraw = 1;
	      /* 再生中フラグをON */
	      engine->playing = 1;
	  }
	} else if (eventAction == AMOTION_EVENT_ACTION_MOVE) {
          /* 画面に指が触れた状態で動かした場合 */
	  /* タップ位置の取得 */
	  float x = AMotionEvent_getX(event, 0); /* X座標 */
	  float y = AMotionEvent_getY(event, 0); /* Y座標 */
	  __android_log_print (ANDROID_LOG_INFO, LOG_TAG,
			       "move = (%f, %f) [%f]", 
			       x, y, x - engine->xpos);
          /*
           * タッチ操作を回転に置き換える
           * 横方向の移動量を Y 軸の回転量に変換
           */
	  engine->yrot = 360.0f / engine->width * (x - engine->xpos);
	  /* 画面を描画するフラグを ON */
	  engine->needToDraw = 1;
	}
        return 1;
    } else if (AInputEvent_getType(event) == AINPUT_EVENT_TYPE_KEY) {
        /* キー入力操作の場合 */
        __android_log_print (ANDROID_LOG_INFO, LOG_TAG,
		"Key event: action=%d keyCode=%d metaState=0x%x",
                AKeyEvent_getAction(event),
                AKeyEvent_getKeyCode(event),
                AKeyEvent_getMetaState(event));
    }

    return 0;
}

/*
 * ウィンドウの初期化、終了、フォーカス等のイベントがあった時に呼び出される関数
 */
static void engine_handle_cmd(struct android_app* app, int32_t cmd)
{
    struct engine* engine = (struct engine*)app->userData;
    switch (cmd) {
        /* ウィンドウが表示される時 */
        case APP_CMD_INIT_WINDOW:
            if (engine->app->window != NULL) {
                engine_init_display(engine);
                engine_draw_frame(engine);
            }
            break;
        /* 終了時 */
        case APP_CMD_TERM_WINDOW:
            engine_term_display(engine);
            break;
        /* アプリのウィンドウにフォーカスがある時 */
        case APP_CMD_GAINED_FOCUS:
            /* センサの値の監視を開始します */
            if (engine->accelerometerSensor != NULL) {
                ASensorEventQueue_enableSensor(engine->sensorEventQueue,
                        engine->accelerometerSensor);
                /* 1 秒あたり 60 回のイベントを受け取れるようにします */
                ASensorEventQueue_setEventRate(engine->sensorEventQueue,
                        engine->accelerometerSensor, (1000L/60)*1000);
            }
            break;
        /* アプリのウィンドウのフォーカスが外れた時 */
        case APP_CMD_LOST_FOCUS:
            /*
             * センサの値の監視を停止します
             * （必要のない時は止めてバッテリの消費を防ぐ）
             */
            if (engine->accelerometerSensor != NULL) {
                ASensorEventQueue_disableSensor(engine->sensorEventQueue,
                        engine->accelerometerSensor);
            }
            engine->needToDraw = 0;
            engine_draw_frame(engine);
            break;
    }
}

/*
 * 音声再生用エンジンの生成
 */
void createSoundEngine()
{
    SLresult result;
    /* エンジンオブジェクトの生成 */
    result = slCreateEngine(&engineObject, 0, NULL, 0, NULL, NULL);
    assert(SL_RESULT_SUCCESS == result);

    /* 生成オブジェクトはリアライズすることで内部リソースが確保されて使用可能になります */
    result = (*engineObject)->Realize(engineObject, SL_BOOLEAN_FALSE);
    assert(SL_RESULT_SUCCESS == result);

    /*
     * エンジンインターフェースの取得
     * 関連する他のオブジェクトを生成するために必要です
     */
    result = (*engineObject)->GetInterface(engineObject, SL_IID_ENGINE, &engineEngine);
    assert(SL_RESULT_SUCCESS == result);

    /* 出力先オブジェクトの生成 */
    result = (*engineEngine)->CreateOutputMix(engineEngine, &outputMixObject, 0, NULL, NULL);
    assert(SL_RESULT_SUCCESS == result);

    /* 生成オブジェクトはリアライズすることで内部リソースが確保されて使用可能になります */
    result = (*outputMixObject)->Realize(outputMixObject, SL_BOOLEAN_FALSE);
    assert(SL_RESULT_SUCCESS == result);
}

/*
 * プレイヤーからのコールバック関数
 * （再生が終了した時に呼び出されます）
 */
void playerCallback(SLPlayItf caller, void *context, SLuint32 event)
{
    struct engine *engine = (struct engine *)context;
    __android_log_print (ANDROID_LOG_INFO, LOG_TAG, "The end of the sound. [%d]", engine->playing);
    /* 再生状態フラグを停止にします */
    engine->playing = 0;
    /* 画面をクリアするために描画必要フラグをONにします */
    engine->needToDraw = 1;
}

/*
 * オーディオプレイヤーオブジェクトの生成
 */
int createAssetAudioPlayer(struct engine *engine)
{
    SLresult result;

    /* AssetManager のインスタンスの取得 */
    AAssetManager *amgr = engine->app->activity->assetManager;
    assert(NULL != amgr);
    /* ファイル名を指定して音声ファイルのassetをオープンします */
    AAsset* asset = AAssetManager_open(amgr, "foo.wav", AASSET_MODE_UNKNOWN);
    // the asset might not be found
    if (NULL == asset) {
        return 0;
    }

    /* オープンしたassetのファイルディスクリプタを取得します */
    off_t start, length;
    int fd = AAsset_openFileDescriptor(asset, &start, &length);
    assert(0 <= fd);
    /* オープンした asset をクローズします */
    AAsset_close(asset);

    /* 音声データの場所を設定 */
    SLDataLocator_AndroidFD loc_fd = {SL_DATALOCATOR_ANDROIDFD, fd, start, length};
    /* 音声データのフォーマット関連の設定 */
    SLDataFormat_MIME format_mime = {SL_DATAFORMAT_MIME, NULL, SL_CONTAINERTYPE_UNSPECIFIED};
    SLDataSource audioSrc = {&loc_fd, &format_mime};

    /* 出力先の設定 */
    SLDataLocator_OutputMix loc_outmix = {SL_DATALOCATOR_OUTPUTMIX, outputMixObject};
    SLDataSink audioSnk = {&loc_outmix, NULL};

    const SLInterfaceID ids[1] = {SL_IID_SEEK};
    const SLboolean req[1] = {SL_BOOLEAN_TRUE};
    /* Audio Player オブジェクトの生成 */
    result = (*engineEngine)->CreateAudioPlayer(engineEngine, &fdPlayerObject, &audioSrc, &audioSnk,
            1, ids, req);
    assert(SL_RESULT_SUCCESS == result);

    /* 生成オブジェクトはリアライズすることで内部リソースが確保されて使用可能になります */
    result = (*fdPlayerObject)->Realize(fdPlayerObject, SL_BOOLEAN_FALSE);
    assert(SL_RESULT_SUCCESS == result);

    /* プレイヤー インターフェースの取得 */
    result = (*fdPlayerObject)->GetInterface(fdPlayerObject, SL_IID_PLAY, &fdPlayerPlay);
    assert(SL_RESULT_SUCCESS == result);

    /* シーク インターフェースの取得 */
    result = (*fdPlayerObject)->GetInterface(fdPlayerObject, SL_IID_SEEK, &fdPlayerSeek);
    assert(SL_RESULT_SUCCESS == result);

    /* プレイヤーインターフェースにコールバック関数を登録します */
    result = (*fdPlayerPlay)->RegisterCallback(fdPlayerPlay, playerCallback, engine);
    assert(SL_RESULT_SUCCESS == result);

    /* コールバックは再生中にファイルの末尾に到達した際に呼ばれるよう設定します */
    result = (*fdPlayerPlay)->SetCallbackEventsMask(fdPlayerPlay,
						    SL_PLAYEVENT_HEADATEND
						    );
    assert(SL_RESULT_SUCCESS == result);

    return 1;
}

/*
 * 音声の再生／停止状態を切り替えます
 * 引数：
 * engine ...... アプリケーション構造体
 * isPlaying ... 再生状態（0: 停止 1: 再生）
 */
void setPlayingAssetAudioPlayer(struct engine *engine, int isPlaying)
{
    SLresult result;

    // make sure the asset audio player was created
    if (NULL != fdPlayerPlay) {
        /* 再生位置を先頭に移動します */
        result = (*(fdPlayerSeek))->SetPosition(fdPlayerSeek, 0, SL_SEEKMODE_FAST);
        assert(SL_RESULT_SUCCESS == result);
        /* 再生状態を設定します */
        result = (*fdPlayerPlay)->SetPlayState(fdPlayerPlay, isPlaying ?
            SL_PLAYSTATE_PLAYING : SL_PLAYSTATE_PAUSED);
        assert(SL_RESULT_SUCCESS == result);
    }
}

/*
 * android_main(struct android_app* state) がエントリの関数となり、
 * アプリが実行された際はここから処理が始まります
 */
void android_main(struct android_app* state)
{
    struct engine engine;

    // Make sure glue isn't stripped.
    app_dummy();

    memset(&engine, 0, sizeof(engine));
    state->userData = &engine;
    /* 引数で受け取った state に対してイベントが発生した際に呼ばれる関数を結びつけています */
    state->onAppCmd = engine_handle_cmd;
    state->onInputEvent = engine_handle_input;
    engine.app = state;

    /* 音声再生用 engine の生成 */
    createSoundEngine();
    /* 音声プレイヤーオブジェクトの生成 */
    createAssetAudioPlayer(&engine);

    /* センサへのアクセスの前準備 */
    /* センサマネージャの取得 */
    engine.sensorManager = ASensorManager_getInstance();
    /* 加速度センサ (accelerometer) の指定 */
    engine.accelerometerSensor = ASensorManager_getDefaultSensor(engine.sensorManager,
            ASENSOR_TYPE_ACCELEROMETER);
    /*
     * イベントキューの生成
     * イベントループの中でセンサからの値の変化を検出できるようにします
     */
    engine.sensorEventQueue = ASensorManager_createEventQueue(engine.sensorManager,
            state->looper, LOOPER_ID_USER, NULL, NULL);

    setIdentityMatrix(engine.matrix);

    /* このアプリケーションのメインループ */
    while (1) {
        int ident;
        int events;
        struct android_poll_source* source;

	/*
	 * ALooper_pollAll() にてイベントのポーリングを行い、
	 * 溜まっているイベント全てを処理するまでループします
	 */
        while ((ident=ALooper_pollAll(engine.needToDraw ? 0 : -1, NULL, &events,
                (void**)&source)) >= 0) {

            // Process this event.
            if (source != NULL) {
                source->process(state, source);
            }

            if (ident == LOOPER_ID_USER) {
	        /* センサーからイベントありの場合 */
                if (engine.accelerometerSensor != NULL) {
                    ASensorEvent event;
                    while (ASensorEventQueue_getEvents(engine.sensorEventQueue,
                            &event, 1) > 0) {
		        /* センサーイベントから加速度センサの値を取り出す */
			engine.acc.x = event.acceleration.x;
			engine.acc.y = event.acceleration.y;
			engine.acc.z = event.acceleration.z;
			/* 値に変化があったので画面を再描画するフラグを ON */
			engine.needToDraw = 1;
                    }
                }
            }

            if (state->destroyRequested != 0) {
                __android_log_print (ANDROID_LOG_INFO, LOG_TAG, "Engine thread destroy requested!");
		/*
		 * 終了要求のイベントがあった場合は終了処理の関数を呼び出します
		 */
                engine_term_display(&engine);
                return;
            }
        }

        if (engine.needToDraw) {
	    /*
	     * 画面の描画が必要と判断された場合は描画関数を呼び出します
	     */
            engine_draw_frame(&engine);
        }
    }
}
